import os # for reading files
import sys
print('[A] Loading ' + sys.argv[0] + '.\n')
import spacy # for nlp
import networkx as nx # for visualisation
import matplotlib.pyplot as plt # for visualisation
from copy import deepcopy
from nltk.corpus import stopwords
import numpy as np # for calculations
nlp = spacy.load('en') # standard english nlp


def frequencies(corpus_path, target, stop_words=[], allowed_tags=['NN','NNS','JJ','JJS','JJR','NNP','VBZ','VBG'], min_context_size = 4, max_nodes=100000, max_edges=10000000):
    
    node_freq = dict()
    edge_freq = dict()
    
    files = [corpus_path+'/'+f for f in os.listdir(corpus_path)]
    s_target = target.replace('_', ' ') #target word with spaces
    
    i = 0
    for f in files[:]:
        
        if i % int(len(files[:])/23) == 0:
            file_ratio = i/len(files[:])
            max_node_ratio = len(node_freq)/max_nodes
            max_edge_ratio = len(edge_freq)/max_edges
            ratios = [file_ratio, max_node_ratio, max_edge_ratio]
            print(' ~{}%\tNodes: {}\tEdges: {}.'.format(int((max(ratios))*100), len(node_freq), len(edge_freq)))
        
        if len(node_freq) > max_nodes:
            return node_freq, edge_freq
        
        if len(edge_freq) > max_edges:
            return node_freq, edge_freq
        
        with open(f, 'r') as lines:
            
            try:
                
                for line in lines:
                    
                    line = line.lower()
                    
                    if s_target in line:
                        
                        tokens = set()
                        doc = nlp(line.replace(s_target, target))
                        
                        if target in [t.text for t in doc]:
                            
                            for tok in doc:
                                
                                text = tok.text
                                tag = tok.tag_
                                
                                if text == target:
                                    pass
                                
                                elif text in stop_words:
                                    pass
                                
                                elif tag in allowed_tags:
                                    tokens.add(tok.text)
                                    
                            if len(tokens) >= min_context_size:
                                for token in tokens:
                                    
                                    if token in node_freq:
                                        node_freq[token] += 1
                                    else:
                                        node_freq[token] = 1
                                
                                for edge in {(x,y) for x in tokens for y in tokens if x < y}:
                                    
                                    if edge in edge_freq:
                                        edge_freq[edge] += 1
                                    else:
                                        edge_freq[edge] = 1
            
            except UnicodeDecodeError:
                
                pass
                #print('Failed to decode:', f)              
        
        i += 1
    
    print(' 100%\tNodes: {}\tEdges: {}.'.format(len(node_freq), len(edge_freq)))
    return node_freq, edge_freq


def build_graph(node_freq, edge_freq, min_node_freq=10, min_edge_freq=5, max_weight= 0.9):
    
    G = nx.Graph()
    
    for key, value in node_freq.items():
        
        if value >= min_node_freq:
            G.add_node(key)
            
    for key, value in edge_freq.items():
        
        if value < min_edge_freq:
            continue
        
        if key[0] not in G.nodes or key[1] not in G.nodes:
            continue
        
        weight = 1 - max(edge_freq[key]/node_freq[key[0]], edge_freq[key]/node_freq[key[1]])
        if weight <= max_weight:
            G.add_edge(*key, weight=weight)
    
    return G


def root_hubs(graph, edge_freq, min_neighbors=4, theshold=0.8):
    
    G = deepcopy(graph)
    V = sorted(G.nodes, key=lambda key: G.degree[key], reverse=True) # -1 to sort descending (...3 -> 2 -> 1...)
    H = list()
    
    while V:
        
        v = V[0]
        
        if G.degree[v] >= min_neighbors:
        
            mfn = sorted(G.adj[v], key=lambda key: edge_freq[v,key] if v < key else edge_freq[key, v], reverse=True)[:min_neighbors] #mfn: most frequent neighbors
            
            if np.mean([G.edges[v,n]['weight'] for n in mfn]) < theshold:
                
                H.append(v)
            
                for nbr in deepcopy(G).adj[v]:
                
                    G.remove_node(nbr)
                
            G.remove_node(v)
            
            V = sorted(G.nodes, key=lambda key: G.degree[key], reverse=True)
        
        else:
        
            return H
    
    return H


#Components algorithm from Véronis (2004), converts graph for target into a MST
def components(graph, hubs, target):
    
    G = deepcopy(graph)
    H = hubs
    t = target
    
    #G.add_node(t)
    #for h in H:
        #G.add_edge(t,h,weight=0)
        
    T = nx.minimum_spanning_tree(G)
    
    for node in deepcopy(T).nodes:
        if len(T.adj[node]) == 0:
            T.remove_node(node)
    
    return T


#Uses MST to disambiguate context, should ideally write to evaluator format
def disambiguate(mst, hubs, contexts):
    
    T = mst
    H = hubs
    C = [c.lower().strip() for c in contexts]
    backup_cluster = len(H)
    result = []
    
    for v in list(T.nodes):
        
        scores = []
        
        for h in H:
            
            if nx.has_path(T,v,h):
                
                path = nx.shortest_path(T,v,h,'weight')
                total_weight = 0
            
                for i in range(1, len(path)):
                    
                    total_weight += T[path[i-1]][path[i]]['weight']
            
                scores.append(1/(1+total_weight))
                
            else:
                
                scores.append(0)
        
        T.nodes[v]['s'] = np.array([s if s == max(scores) else 0 for s in scores])
    
    for c in C:
        
        toks = [t.text for t in nlp(c)]
        vector = np.sum([T.nodes[t]['s'] if t in T.nodes else np.zeros(len(H)) for t in toks], axis=0)
        
        idx = C.index(c) + 1
        
        if len(vector) == 0: #if no senses are found -> all in one
            result.append((0, idx))
        elif max(vector) == 0: #if no sense matches -> singletons
            pass
        else: 
            result.append((np.argmax(vector), idx))
        
    return result


def backup(contexts):
    
    pass
        
        

if __name__ == '__main__':
    
    data_path = '/home/students/zimmermann/Courses/ws17/fsem/absinth/WSI-Evaluator/datasets/MORESQUE/'
    #corpus_path = '/home/students/zimmermann/Courses/ws17/fsem/absinth/test'
    corpus_path = '/proj/absinth/wikipedia.txt.dump.20140615-en.SZTAKI/'
    results_path = '/home/students/zimmermann/Courses/ws17/fsem/absinth/clustering/'
    
    stop = set(stopwords.words('english') + ['utc', 'new', 'other', 'talk', 'wikipedia', 'article', 'topic', 'page', 'editors', 'encyclopedia', 'free'])
    
    results = dict()
    
    with open(data_path+'results.txt', 'r') as results_file:
        
        for line in results_file.readlines()[1:]:
            
            l = line.split('\t')
            id1, _ = l[0].split('.')
            
            if id1 not in results:
                results[id1]=list()
                
            results[id1].append(" ".join(l[2:]))
            
            
    topics = dict()
    
    with open(data_path+'topics.txt', 'r') as topics_file:
        
        already_processed = [f.replace('.absinth', '') for f in os.listdir(results_path)]
        
        for line in topics_file.readlines()[1:]:
            
            l = line.split('\t')
            if l[1] not in already_processed:
                topics[l[0]] = l[1]
        
    for key, value in topics.items():
            
        o_target = value.strip() #original target
        print("[A] Processing '"+o_target+"'.\n")
        if o_target[:4] == 'the_' and o_target.count('_') >= 2: #hard coded 'the'-protection
            target = o_target[4:]
        else:
            target = o_target
        
        f = open(results_path+target+'.absinth', 'w')
        f.write('subTopicID\tresultID\n')
        
        print('[A] Counting Tokens...')
        node_freq, edge_freq = frequencies(corpus_path, target, stop)
        
        print('\n[A] Building Graph.\n')
        G = build_graph(node_freq, edge_freq)
        
        print('[A] Collecting Root Hubs...')
        H = root_hubs(G, edge_freq)
        for h in H:
            mfn = sorted(G.adj[h], key=lambda key: edge_freq[h,key] if h < key else edge_freq[key, h], reverse=True)[:6]
            print(' {}: {}\n'.format(h, mfn))
        
        print('[A] Building Minimum Spanning Tree.\n')
        T = components(G, H, target)
    
        print('[A] Disambiguating Results...')
        D = disambiguate(T, H, results[key])
        print(' Mapping:', D, '\n')
        
        print('[A] Writing to file '+o_target+'.absinth.\n\n')
        for d in D:
            
            f.write(key+'.'+str(d[0])+'\t'+key+'.'+str(d[1])+'\n')
            
        f.close()
    
    #nx.draw(T, with_labels=True)
    #plt.show()
